package io.tswf.transmutations.common.reader

import groovy.transform.CompileStatic
import io.tswf.transmutations.common.annotation.Primitive
import io.tswf.transmutations.common.exception.InvalidClassReferenceAnnotationException
import org.objectweb.asm.Type
import org.objectweb.asm.tree.AnnotationNode

@CompileStatic
class ClassReferenceReader {

    public static final String CLASS_DESCRIPTOR_PARAMETER_NAME = "descriptor"
    public static final String CLASS_PARAMETER_NAME = "value"
    public static final String PRIMITIVE_PARAMETER_NAME = "primitive"

    public static final Set<String> PARAMETER_NAMES = [
            CLASS_DESCRIPTOR_PARAMETER_NAME,
            CLASS_PARAMETER_NAME,
            PRIMITIVE_PARAMETER_NAME
    ] as Set

    static Type getTypeFromClassRefAnnotation(AnnotationNode annotationNode) {
        findTypeFromClassRefAnnotation(annotationNode)
                ?.ensureNotNull { new InvalidClassReferenceAnnotationException("There is no class reference parameters on ${annotationNode.desc}!") }
    }

    static Type findTypeFromClassRefAnnotation(AnnotationNode annotationNode) {
        def descriptorValue = annotationNode.findValueByName(CLASS_DESCRIPTOR_PARAMETER_NAME, String)
        def classValue = annotationNode.findValueByName(CLASS_PARAMETER_NAME, Type)
        def primitiveEnumValue = annotationNode.findValueByName(PRIMITIVE_PARAMETER_NAME, Primitive)

        def presentedParameters = PARAMETER_NAMES
            .stream()
                .filter { annotationNode.findValueByName(it) != null }
                .toSet()

        if (presentedParameters.size() > 1) {
            throw new InvalidClassReferenceAnnotationException("The class reference annotation '${annotationNode.type.className}' must contain only one type parameter, but found ${presentedParameters} presented parameters that can not be used together")
        }

        if (presentedParameters.isEmpty()) {
            throw new InvalidClassReferenceAnnotationException("The class reference annotation '${annotationNode.type.className}' must contain only one type parameter, but found, but no one found!")
        }

        if (primitiveEnumValue != null) {
            return Type.getType(primitiveEnumValue.javaClass)
        }

        if (descriptorValue != null) {
            return Type.getType(descriptorValue)
        }

        if (classValue != null) {
            return classValue
        }

        return null
    }
}
